@@ -8,11 +8,13 @@ from dotenv import load_dotenv
 import pytest
 import responses
 import requests
+from redbird.repos.csv import CSVFileRepo
+from redbird.repos.json import JSONDirectoryRepo
 from redbird.repos.rest import RESTRepo
 from redbird.repos.sqlalchemy import SQLRepo
 from redbird.repos.memory import MemoryRepo
 from redbird.repos.mongo import MongoRepo
-from redbird.oper import greater_equal, greater_than, less_equal, less_than, not_equal
+from redbird.oper import skip, between, greater_equal, greater_than, less_equal, less_than, not_equal
 
 from pydantic import BaseModel, Field
 
@@ -27,9 +29,14 @@ TEST_CASES = [
     pytest.param("mongo"),
     pytest.param("mongo-mock"),
     pytest.param("http-rest"),
+    pytest.param("csv"),
+    pytest.param("json-dir"),
 ]
 
 
+def sort_items(items, repo, field="id"):
+    return list(sorted(items, key=lambda x: repo.get_field_value(x, field)))
+
 # ------------------------
 # ACTUAL TESTS
 # ------------------------
@@ -104,49 +111,90 @@ class TestPopulated(RepoTests):
     def test_filter_by_first(self, repo):
         self.populate(repo)
         Item = repo.model
-        assert repo.filter_by(age=30).first() == Item(id="b", name="John", age=30)
+        if repo.ordered:
+            assert repo.filter_by(age=30).first() == Item(id="b", name="John", age=30)
+        else:
+            item = repo.filter_by(age=30).first()
+            assert repo.get_field_value(item, 'age') == 30
 
     def test_filter_by_last(self, repo):
         self.populate(repo)
         Item = repo.model
-        assert repo.filter_by(age=30).last() == Item(id="d", name="Johnny", age=30)
+        if repo.ordered:
+            assert repo.filter_by(age=30).last() == Item(id="d", name="Johnny", age=30)
+        else:
+            item = repo.filter_by(age=30).last()
+            assert repo.get_field_value(item, 'age') == 30
 
     def test_filter_by_limit(self, repo):
         self.populate(repo)
         Item = repo.model
-        assert repo.filter_by(age=30).limit(2) == [
-            Item(id="b", name="John", age=30),
-            Item(id="c", name="James", age=30),
-        ]
+
+        actual = repo.filter_by(age=30).limit(2)
+        expected = [
+                Item(id="b", name="John", age=30),
+                Item(id="c", name="James", age=30),
+            ]
+        if not repo.ordered:
+            # We don't know which are the top 2
+            assert isinstance(actual, list)
+            for item in actual:
+                assert repo.get_field_value(item, "age") == 30
+            assert len(actual) == 2
+        else:
+            assert actual == expected
 
     def test_filter_by_all(self, repo):
         self.populate(repo)
         Item = repo.model
-        assert repo.filter_by(age=30).all() == [
+        
+        actual = repo.filter_by(age=30).all()
+        expected = [
             Item(id="b", name="John", age=30),
             Item(id="c", name="James", age=30),
             Item(id="d", name="Johnny", age=30),
         ]
+        
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_filter_by_update(self, repo):
         self.populate(repo)
         Item = repo.model
 
         repo.filter_by(age=30).update(name="Something")
-        assert repo.filter_by().all() == [
+
+        actual = repo.filter_by().all()
+        expected = [
             Item(id="a", name="Jack", age=20),
             Item(id="b", name="Something", age=30),
             Item(id="c", name="Something", age=30),
             Item(id="d", name="Something", age=30),
             Item(id="e", name="Jesse", age=40),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_filter_by_replace(self, repo):
         self.populate(repo)
         Item = repo.model
 
         repo.filter_by(id="b").replace(id="b", name="Something")
+
+        expected = [
+            Item(id="a", name="Jack", age=20),
+            Item(id="c", name="James", age=30),
+            Item(id="d", name="Johnny", age=30),
+            Item(id="e", name="Jesse", age=40),
+            Item(id="b", name="Something", age=None),
+        ]
         if repo.model == dict and not isinstance(repo, SQLRepo):
+            # In dict repositories the not given fields don't exists
+            # (exception: structured data stores)
             expected = [
                 Item(id="a", name="Jack", age=20),
                 Item(id="c", name="James", age=30),
@@ -154,15 +202,13 @@ class TestPopulated(RepoTests):
                 Item(id="e", name="Jesse", age=40),
                 Item(id="b", name="Something"),
             ]
-        else:
-            expected = [
-                Item(id="a", name="Jack", age=20),
-                Item(id="c", name="James", age=30),
-                Item(id="d", name="Johnny", age=30),
-                Item(id="e", name="Jesse", age=40),
-                Item(id="b", name="Something", age=None),
-            ]
-        assert repo.filter_by().all() == expected
+
+        actual = repo.filter_by().all()
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+
+        assert actual == expected
 
     def test_filter_by_replace_dict(self, repo):
         self.populate(repo)
@@ -170,7 +216,16 @@ class TestPopulated(RepoTests):
 
         repo.filter_by(id="b").replace({"id": "b", "name": "Something"})
         
+        expected = [
+            Item(id="a", name="Jack", age=20),
+            Item(id="c", name="James", age=30),
+            Item(id="d", name="Johnny", age=30),
+            Item(id="e", name="Jesse", age=40),
+            Item(id="b", name="Something", age=None),
+        ]
         if repo.model == dict and not isinstance(repo, SQLRepo):
+            # In dict repositories the not given fields don't exists
+            # (exception: structured data stores)
             expected = [
                 Item(id="a", name="Jack", age=20),
                 Item(id="c", name="James", age=30),
@@ -178,29 +233,32 @@ class TestPopulated(RepoTests):
                 Item(id="e", name="Jesse", age=40),
                 Item(id="b", name="Something"),
             ]
-        else:
-            expected = [
-                Item(id="a", name="Jack", age=20),
-                Item(id="c", name="James", age=30),
-                Item(id="d", name="Johnny", age=30),
-                Item(id="e", name="Jesse", age=40),
-                Item(id="b", name="Something", age=None),
-            ]
 
-        assert repo.filter_by().all() == expected
+        actual = repo.filter_by().all()
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+
+        assert actual == expected
 
     def test_filter_by_delete(self, repo):
         self.populate(repo)
         Item = repo.model
 
         repo.filter_by(age=30).delete()
-        assert repo.filter_by().all() == [
+
+        actual = repo.filter_by().all()
+        expected = [
             Item(id="a", name="Jack", age=20),
             #Item(id="b", name="John", age=30),
             #Item(id="c", name="James", age=30),
             #Item(id="d", name="Johnny", age=30),
             Item(id="e", name="Jesse", age=40),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_filter_by_count(self, repo):
         self.populate(repo)
@@ -214,13 +272,18 @@ class TestPopulated(RepoTests):
             pytest.xfail("RESTRepo does not support operations (yet)")
 
         Item = repo.model
-        assert repo.filter_by(age=less_than(40)).all() == [
+        actual = repo.filter_by(age=less_than(40)).all() 
+        expected = [
             Item(id="a", name="Jack", age=20),
             Item(id="b", name="John", age=30),
             Item(id="c", name="James", age=30),
             Item(id="d", name="Johnny", age=30),
             #Item(id="e", name="Jesse", age=40),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_getitem(self, repo):
         self.populate(repo)
@@ -238,13 +301,20 @@ class TestPopulated(RepoTests):
         Item = repo.model
 
         del repo["b"]
-        assert repo.filter_by().all() == [
+
+
+        actual = repo.filter_by().all() 
+        expected = [
             Item(id="a", name="Jack", age=20),
             #Item(id="b", name="John", age=30),
             Item(id="c", name="James", age=30),
             Item(id="d", name="Johnny", age=30),
             Item(id="e", name="Jesse", age=40),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     @pytest.mark.skip(reason="This is a minor issue")
     def test_delitem_missing(self, repo):
@@ -258,13 +328,18 @@ class TestPopulated(RepoTests):
 
         repo["d"] = {"name": "Johnny boy"}
 
-        assert repo.filter_by().all() == [
+        actual = repo.filter_by().all() 
+        expected = [
             Item(id="a", name="Jack", age=20),
             Item(id="b", name="John", age=30),
             Item(id="c", name="James", age=30),
             Item(id="d", name="Johnny boy", age=30),
             Item(id="e", name="Jesse", age=40),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     @pytest.mark.skip(reason="This is a minor issue")
     def test_setitem_missing(self, repo):
@@ -279,6 +354,20 @@ class TestPopulated(RepoTests):
 )
 class TestFilteringOperations(RepoTests):
 
+    def populate(self, repo, items=None):
+        if items is None:
+            items = [
+                dict(id="a", name="Jack", age=20),
+                dict(id="b", name="John", age=30),
+                dict(id="c", name="James", age=30),
+                dict(id="d", name="Johnny", age=30),
+                dict(id="e", name="Jesse", age=40),
+                dict(id="f", name="Jim", age=41),
+            ]
+        for d in items:
+            item = repo.to_item(d)
+            repo.add(item)
+
     def test_greater_than(self, repo):
         self.populate(repo)
 
@@ -286,13 +375,20 @@ class TestFilteringOperations(RepoTests):
             pytest.xfail("RESTRepo does not support operations (yet)")
 
         Item = repo.model
-        assert repo.filter_by(age=greater_than(20)).all() == [
+
+        actual = repo.filter_by(age=greater_than(20)).all()
+        expected = [
             #Item(id="a", name="Jack", age=20),
             Item(id="b", name="John", age=30),
             Item(id="c", name="James", age=30),
             Item(id="d", name="Johnny", age=30),
             Item(id="e", name="Jesse", age=40),
+            Item(id="f", name="Jim", age=41),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_less_than(self, repo):
         self.populate(repo)
@@ -301,26 +397,37 @@ class TestFilteringOperations(RepoTests):
             pytest.xfail("RESTRepo does not support operations (yet)")
     
         Item = repo.model
-        assert repo.filter_by(age=less_than(30)).all() == [
+        actual = repo.filter_by(age=less_than(30)).all()
+        expected = [
             Item(id="a", name="Jack", age=20),
             #Item(id="b", name="John", age=30),
             #Item(id="c", name="James", age=30),
             #Item(id="d", name="Johnny", age=30),
             #Item(id="e", name="Jesse", age=40),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_greater_equal(self, repo):
         self.populate(repo)
         if isinstance(repo, RESTRepo):
             pytest.xfail("RESTRepo does not support operations (yet)")
         Item = repo.model
-        assert repo.filter_by(age=greater_equal(30)).all() == [
+        actual = repo.filter_by(age=greater_equal(30)).all()
+        expected = [
             #Item(id="a", name="Jack", age=20),
             Item(id="b", name="John", age=30),
             Item(id="c", name="James", age=30),
             Item(id="d", name="Johnny", age=30),
             Item(id="e", name="Jesse", age=40),
+            Item(id="f", name="Jim", age=41),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_less_equal(self, repo):
         self.populate(repo)
@@ -329,13 +436,18 @@ class TestFilteringOperations(RepoTests):
             pytest.xfail("RESTRepo does not support operations (yet)")
 
         Item = repo.model
-        assert repo.filter_by(age=less_equal(30)).all() == [
+        actual = repo.filter_by(age=less_equal(30)).all()
+        expected = [
             Item(id="a", name="Jack", age=20),
             Item(id="b", name="John", age=30),
             Item(id="c", name="James", age=30),
             Item(id="d", name="Johnny", age=30),
             #Item(id="e", name="Jesse", age=40),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_not_equal(self, repo):
         self.populate(repo)
@@ -344,13 +456,57 @@ class TestFilteringOperations(RepoTests):
             pytest.xfail("RESTRepo does not support operations (yet)")
 
         Item = repo.model
-        assert repo.filter_by(age=not_equal(30)).all() == [
+        actual = repo.filter_by(age=not_equal(30)).all()
+        expected = [
             Item(id="a", name="Jack", age=20),
             #Item(id="b", name="John", age=30),
             #Item(id="c", name="James", age=30),
             #Item(id="d", name="Johnny", age=30),
             Item(id="e", name="Jesse", age=40),
+            Item(id="f", name="Jim", age=41),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
+
+    def test_between(self, repo):
+        self.populate(repo)
+        if isinstance(repo, RESTRepo):
+            pytest.xfail("RESTRepo does not support operations (yet)")
+        Item = repo.model
+        actual = repo.filter_by(age=between(30, 40)).all() 
+        expected = [
+            #Item(id="a", name="Jack", age=20),
+            Item(id="b", name="John", age=30),
+            Item(id="c", name="James", age=30),
+            Item(id="d", name="Johnny", age=30),
+            Item(id="e", name="Jesse", age=40),
+            #Item(id="e", name="Jim", age=41),
+        ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
+
+    def test_skip(self, repo):
+        self.populate(repo)
+        if isinstance(repo, RESTRepo):
+            pytest.xfail("RESTRepo does not support operations (yet)")
+        Item = repo.model
+        actual = repo.filter_by(age=skip).all()
+        expected = [
+            Item(id="a", name="Jack", age=20),
+            Item(id="b", name="John", age=30),
+            Item(id="c", name="James", age=30),
+            Item(id="d", name="Johnny", age=30),
+            Item(id="e", name="Jesse", age=40),
+            Item(id="f", name="Jim", age=41),
+        ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
 @pytest.mark.parametrize(
     'repo',
@@ -368,10 +524,16 @@ class TestEmpty:
         assert repo.filter_by().all() == [Item(id="a", name="Jack", age=20)]
 
         repo.add(Item(id="b", name="John", age=30))
-        assert repo.filter_by().all() == [
+        
+        actual = repo.filter_by().all()
+        expected = [
             Item(id="a", name="Jack", age=20),
             Item(id="b", name="John", age=30)
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_update(self, repo):
         Item = repo.model
@@ -380,11 +542,17 @@ class TestEmpty:
         repo.add(Item(id="b", name="John", age=30))
 
         repo.update(Item(id="a", name="Max"))
-        items = repo.filter_by().all()
-        assert items == [
+        actual = repo.filter_by().all()
+
+        actual = repo.filter_by().all()
+        expected = [
             Item(id="a", name="Max", age=20),
             Item(id="b", name="John", age=30),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_replace(self, repo):
         Item = repo.model
@@ -405,8 +573,10 @@ class TestEmpty:
                 Item(id="b", name="John", age=30),
             ]
 
-        items = sorted(repo.filter_by().all(), key=lambda x: repo.get_field_value(x, "id"))
-        assert items == expected
+        actual = repo.filter_by().all()
+        actual = sort_items(actual, repo)
+        expected = sort_items(expected, repo)
+        assert actual == expected
 
     def test_add_exist(self, repo):
         Item = repo.model
@@ -442,16 +612,29 @@ class TestEmpty:
 
         repo.add(Item(id="a", name="Max", age=20))
         repo.add(Item(id="b", name="John", age=30))
-        assert list(repo) == [
+
+        # asserting
+        actual = list(repo)
+        expected =[
             Item(id="a", name="Max", age=20),
             Item(id="b", name="John", age=30),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
         repo.get_by("a").update(age=40)
-        assert list(repo) == [
+
+        actual = list(repo)
+        expected = [
             Item(id="a", name="Max", age=40),
             Item(id="b", name="John", age=30),
         ]
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
 
         repo.get_by("b").delete()
         assert list(repo) == [
@@ -522,49 +705,13 @@ class TestMalformedData(RepoTests):
         repo.errors_query = "discard"
 
         Item = MalformedItem
-        assert repo.filter_by().all() == [
+        actual = repo.filter_by().all() 
+        expected = [
             Item(id="a", name=1, age=20),
             Item(id="d", name=2, age=30),
             Item(id="e", name=3, age=30),
         ]
-
-    def test_validate_success(self, repo):
-
-        self.populate(repo, [
-            {"id": "a", "name": 1, "age": 20},
-            {"id": "b", "name": "Jack", "age": 30},
-            {"id": "c", "name": "James", "age": 40},
-        ])
-        repo.filter_by().validate()
-
-    def test_validate_fail(self, repo):
-
-        class MalformedItem(BaseModel):
-            id: str
-            name: int
-            age: int
-
-        self.populate(repo, [
-            {"id": "a", "name": 1, "age": 20},
-            {"id": "b", "name": "Jack", "age": 30},
-            {"id": "c", "name": "James", "age": 40},
-        ])
-        repo.model = MalformedItem
-        with pytest.raises(ValueError):
-            repo.filter_by().validate()
-
-    def test_validate_fail_max_shown(self, repo):
-
-        class MalformedItem(BaseModel):
-            id: str
-            name: int
-            age: int
-
-        self.populate(repo, [
-            {"id": "a", "name": 1, "age": 20},
-            {"id": "b", "name": "Jack", "age": 30},
-            {"id": "c", "name": "James", "age": 40},
-        ])
-        repo.model = MalformedItem
-        with pytest.raises(ValueError):
-            repo.filter_by().validate(max_shown=1)
\ No newline at end of file
+        if not repo.ordered:
+            actual = sort_items(actual, repo)
+            expected = sort_items(expected, repo)
+        assert actual == expected
\ No newline at end of file
